import pandas as pd
from .sort import sortPed

def selectPed(data,id,generation=3):###缺失值用0表示
    '''Select pedigree.

    data: the whole pedigree with dataframe type.
    id: a list containing id which want to be selected.
    generation: int value.  Default value is 3. 
    '''
    if not isinstance(id, list): ###必须是list
        print("Please provide id with list type!")
        return
    if not isinstance(generation, int):
        print("ERROR: Parameter generation should be int type!")
        return
    if  generation <=0:
        print("ERROR: Parameter generation should be much than 0")
        return
    ped_ord = sortPed(data)
    
    ###判断id里面提供的个体是不是都在ped里面
    id = pd.Series(id).astype(str)
    if not all(id.isin(ped_ord['id'])):
        print("ERROR: there are some id not in the pedigree.")
        return
    i = 1
    ped_select = pd.DataFrame({'id':[],'sire':[],'dam':[]})
    ped_select =pd.concat([ped_select,ped_ord[ped_ord['id'].isin(id)]])
    while i < generation:
        
        i += 1
        ##去除重复和0
        id = pd.concat([ped_select['sire'],ped_select['dam']])
        id = id.drop_duplicates()
        id = id[id != '0']
        ped_select = pd.concat([ped_select,ped_ord[ped_ord['id'].isin(id)]])
        ped_select.drop_duplicates(inplace=True)
    
    ped_select = sortPed(ped_select)
    #print(ped_select)
    return ped_select


from scipy.cluster import hierarchy
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.patches as pt
import pandas as pd
import numpy as np
from .sort import sortPed
def cluster(A):
    '''Plot the cluster of kinship matrix.

    A: a list with two elements, A[0] is kinship matrix; A[1] is id series of all individuals.
    '''
    if not isinstance(A, list):
        print("ERROR: Parameter A should be a list!")
        return
    if len(A) != 2:
        print("ERROR: A should have 2 elements")
        return
    if not isinstance(A[1], pd.Series):
        print("ERROR: A should have 2 elements with numpy ndarray and pandas Series")
        return
    if not isinstance(A[0], np.ndarray):
        print("ERROR: A should have 2 elements with numpy ndarray and pandas Series")
        return
    if np.isnan(A[0]).any():
        print("ERROR: Nan in A")
        return
    if A[0].shape[0] != len(A[1]):
        print("ERROR: The dimension of A is not equal to the number of individual with genotyped")
        return
    Z = hierarchy.linkage(A[0], method ='average',metric='euclidean')
    P = hierarchy.dendrogram(Z,labels=list(A[1]))
    return P

def pca(A,group):
    '''Plot the PCA of kinship matrix.

    A: a list with two elements, A[0] is kinship matrix; A[1] is id series of all individuals.
    group: a list containing information about the group of individuals.
    '''
    if not isinstance(A, list):
        print("ERROR: Parameter A should be a list!")
        return
    if len(A) != 2:
        print("ERROR: A should have 2 elements")
        return
    if not isinstance(A[1], pd.Series):
        print("ERROR: A should have 2 elements with numpy ndarray and pandas Series")
        return
    if not isinstance(A[0], np.ndarray):
        print("ERROR: A should have 2 elements with numpy ndarray and pandas Series")
        return
    if np.isnan(A[0]).any():
        print("ERROR: Nan in A")
        return
    if A[0].shape[0] != len(A[1]):
        print("ERROR: The dimension of A is not equal to the number of individual with genotyped")
        return
    if len(group) != len(A[1]):
        print("ERROR: The dimension of A is not equal to the length of group.")
        return
    [eigValue, V] = np.linalg.eig(A[0])
    index = np.argsort(-eigValue)
    V = V[:,index]
    eigValue = eigValue[index]
    explained_variance = eigValue / (A[0].shape[0])   # Get variance explained by singular values
    total_var = explained_variance.sum()
    explained_variance_ratio = explained_variance / total_var 
    var=np.cumsum(np.round(explained_variance_ratio, decimals=3)*100)
    

    plt.figure(figsize=(16,7))
    ax = plt.subplot(121)
    ax.plot(list(range(1,len(var)+1)), var)
    ax.set_title("PCA Variance Explained Based on Custom PCA")
    ax.set_ylabel("% Variance Explained")
    ax.set_xlabel("# Of Features")

    pcs = pd.DataFrame({"group":group,
    "PC1":V[:,0],
    "PC2":V[:,1]
            })
    levels, categories = pd.factorize(pcs['group'])
    colors = [plt.cm.tab10(i) for i in levels]  ###暂时选择10种默认颜色，可能不够用
    ax = plt.subplot(122)
    ax.scatter(pcs['PC1'], pcs['PC2'], c=colors)
    x_label = 'PC1(%s%%)' % round((explained_variance_ratio[0]*100.0),2)   #x轴标签字符串
    y_label = 'PC2(%s%%)' % round((explained_variance_ratio[1]*100.0),2)   #y轴标签字符串
    ax.set_xlabel(x_label)    #绘制x轴标签
    ax.set_ylabel(y_label)    #绘制y轴标签
    ax.set_title("PCA plot of top 2 PCs")

    handles = [pt.Patch(color=plt.cm.tab10(i), label=c) for i, c in enumerate(categories)]
    ax.legend(handles=handles,  title='Group')
    return ax

def heat(A):
    '''Plot the heatmap of kinship matrix.

    A: a list with two elements, A[0] is kinship matrix; A[1] is id series of all individuals.
    '''
    if not isinstance(A, list):
        print("ERROR: Parameter A should be a list!")
        return
    if len(A) != 2:
        print("ERROR: A should have 2 elements")
        return
    if not isinstance(A[1], pd.Series):
        print("ERROR: A should have 2 elements with numpy ndarray and pandas Series")
        return
    if not isinstance(A[0], np.ndarray):
        print("ERROR: A should have 2 elements with numpy ndarray and pandas Series")
        return
    if np.isnan(A[0]).any():
        print("ERROR: Nan in A")
        return
    if A[0].shape[0] != len(A[1]):
        print("ERROR: The dimension of A is not equal to the number of individual with genotyped")
        return
    fig = plt.figure()
    # 定义画布为1*1个划分，并在第1个位置上进行作图
    ax = fig.add_subplot(111)
    # 定义横纵坐标的刻度
    ax.set_yticks(range(len(A[1])))
    ax.set_yticklabels(A[1])
    ax.set_xticks(range(len(A[1])))
    ax.set_xticklabels(A[1])
    # 作图并选择热图的颜色填充风格，这里选择hot
    im = ax.imshow(A[0], cmap=plt.cm.summer)
    # 增加右侧的颜色刻度条
    plt.colorbar(im)
    # 增加标题
    plt.title("Heatmap of relationship matrix")

    return ax
def gragh(data_ord):
    '''Plot family tree in three generations of a individual.

    data_ord: the pedigree of one individual in three generations.
    '''
    data_ord = sortPed(data_ord)
    ###如果长度超过15
    if data_ord.shape[0]>15:
        print("Error: Please provide an individual's three-generation pedigree, which should not exceed 15 records at most")
        print("You can selectPed function first to select ped")
        return
    ##先找到目标个体
    plotid = data_ord[~data_ord.iloc[:,0].isin(pd.concat([data_ord.iloc[:,1],(data_ord.iloc[:,2])]).to_list())]
    if plotid.shape[0] != 1:
        print("Error: Please provide one individual's three-generation pedigree, which should not exceed 15 records at most")
        print("You can selectPed function first to select ped")
        return
    id_15 = plotid.iloc[0,0]
    id_7 = plotid.iloc[0,1]
    id_14 = plotid.iloc[0,2]
    if id_7 == '0':
        id_3 = id_6 = id_1 =id_2 =id_4 = id_5 ='0'
    else:
        id_3 = data_ord[data_ord.iloc[:,0] == id_7].iloc[0,1]
        id_6 = data_ord[data_ord.iloc[:,0] == id_7].iloc[0,2]
        if id_3 == '0':
            id_1 =id_2 ='0'
        else:
            id_1 =data_ord[data_ord.iloc[:,0] == id_3].iloc[0,1]
            id_2 =data_ord[data_ord.iloc[:,0] == id_3].iloc[0,2]
        if id_6 =='0':
            id_4 = id_5 ='0'
        else:
            id_4 =data_ord[data_ord.iloc[:,0] == id_6].iloc[0,1]
            id_5 =data_ord[data_ord.iloc[:,0] == id_6].iloc[0,2]
    if id_14 == '0':
        id_8 = id_9 = id_10 =id_11 =id_12 = id_13 ='0'
    else:
        id_10 = data_ord[data_ord.iloc[:,0] == id_14].iloc[0,1]
        id_13 = data_ord[data_ord.iloc[:,0] == id_14].iloc[0,2]
        if id_10 =='0':
            id_8 = id_9 ='0'
        else:
            id_8 = data_ord[data_ord.iloc[:,0] == id_10].iloc[0,1]
            id_9 = data_ord[data_ord.iloc[:,0] == id_10].iloc[0,2]
        if id_13 =='0':
            id_11 = id_12 ='0'
        else:
            id_11 = data_ord[data_ord.iloc[:,0] == id_13].iloc[0,1]
            id_12 = data_ord[data_ord.iloc[:,0] == id_13].iloc[0,2]
    a ='''digraph G {
  edge [dir=none];
  node [shape=box];
  graph [splines=ortho];
 
  1     [shape=box, label="%s",regular=0, color="blue", style="filled" fillcolor="lightblue"] ;
  2     [shape=oval,label="%s", regular=0, color="red", style="filled" fillcolor="pink"] ;
  3     [group =g1,label="%s",shape=box, regular=0, color="blue", style="filled" fillcolor="lightblue"] ;
  4    [shape=box, label="%s",regular=0, color="blue", style="filled" fillcolor="lightblue"] ;
  5   [shape=oval, label="%s",regular=0, color="red", style="filled" fillcolor="pink"] ;
  6   [group =g2,label="%s",shape=oval, regular=0, color="red", style="filled" fillcolor="pink"] ;
  7      [group =g3,label="%s",shape=box, regular=0, color="blue", style="filled" fillcolor="lightblue"] ;
  8     [shape=box, label="%s",regular=0, color="blue", style="filled" fillcolor="lightblue"] ;
  9     [shape=oval, label="%s",regular=0, color="red", style="filled" fillcolor="pink"] ;
  10      [group =b1,label="%s",shape=box, regular=0, color="blue", style="filled" fillcolor="lightblue"] ;
  11     [shape=box,label="%s", regular=0, color="blue", style="filled" fillcolor="lightblue"] ;
  12    [shape=oval,label="%s", regular=0, color="red", style="filled" fillcolor="pink"] ;
  13     [group =b2,label="%s",shape=oval, regular=0, color="red", style="filled" fillcolor="pink"] ;
  14    [group =b3,label="%s",shape=oval, regular=0, color="red", style="filled" fillcolor="pink"] ;
  15   [group =c1,label="%s",shape=box, regular=0, color="blue", style="filled" fillcolor="lightblue"] ;


  a1 [shape=diamond,label="",height=0.25,width=0.25,group =g1];
  {rank=same; 1 -> a1 -> 2};
  a1 -> 3
  a2 [shape=diamond,label="",height=0.25,width=0.25,group =g2];
  {rank=same; 4 -> a2 -> 5};
  a2-> 6
  t1 [style=invis,shape=point,width=0.0,group =g3];
  {edge[ style=invis];rank = same; 2->t1->4;}

  a3 [shape=diamond,label="",height=0.25,width=0.25,group =g3];
  {rank=same; 3 -> a3 -> 6};
  
  a3-> 7
  {edge[style=invis]; t1->a3;}

  b1 [shape=diamond,label="",height=0.25,width=0.25,group =b1];
  {rank=same; 8 -> b1 -> 9};
  b1 -> 10
  b2 [shape=diamond,label="",height=0.25,width=0.25,group =b2];
  {rank=same; 11 -> b2 -> 12};
  b2 -> 13

  t2 [style=invis,shape = point,width=0.1,group =b3];
  {edge[ style=invis];rank = same; 9->t2->11;}

  b3 [shape=diamond,label="",height=0.25,width=0.25,group =b3];
  {rank=same; 10 -> b3 -> 13};
  
  b3 -> 14
  {edge[style=invis]; t2->b3;}
  c1 [shape=diamond,label="",height=0.25,width=0.25,group =c1];
  {rank=same; 7 -> c1 -> 14};
  c1 -> 15

  
  {
  rank = same;
  edge[ style=invis];
  3 -> 6 -> 10 -> 13 ;
  rankdir = LR;
  }

}
  ''' %(id_1,id_2,id_3,id_4,id_5,id_6,id_7,id_8,id_9,id_10,id_11,id_12,id_13,id_14,id_15)
    return a


#Ross, E. M., Moate, P. J., Marett, L. C., Cocks, B. G., & Hayes, B. J. (2013). Metagenomic predictions: From microbiome to complex health and environmental phenotypes in humans and cattle. 
import numpy as np
def makeM(otu):
    '''Calculate the kinship matrix using OTU.
    
    otu: OTU data in numpy ndarray type. Rows are individuals and columns are OTUs.
    '''
    ###检测没有空值
    if not isinstance(otu, np.ndarray):
        print("ERROR: OTU matrix should be numpy ndarray")
        return
    if np.isnan(otu).any():
        print("ERROR: Nan in OTU")
        return
    otu_log = np.log(otu+0.001)
    sd_logSj  = np.std(otu_log,axis=0)
    mean_logSj = np.mean(otu_log,axis=0)
    X =(otu_log-mean_logSj)/sd_logSj
    O = (1/otu.shape[1])*X.dot(X.T)
    return O


from .amatrix import makeA, makeD
from .gmatrix import makeG
from .hmatrix import makeH
from .sort import sortPed
from .loaddata import loadEgPed, loadEgGeno
from .selectped import selectPed
from .plot import cluster, heat,gragh, pca
from .coef import coefKinship,coefInbreeding
from .inter_matrix import makeG_inter
from .mmatrix import makeM


import pandas as pd
import ctypes 
import numpy as np
from scipy import sparse
from scipy.sparse.linalg import spsolve_triangular
import gc
import FUNC

from multiprocessing import Pool
def makeA(data_ord,Sparse=False):
    '''Calculate the additive kinship matrix using pedigree.

    data_ord: a dataframe of pedigree after sort.
    Sparse: bool value. Default value is False. Using sparse matrix can save memory but consume more time.
    '''
    if not isinstance(data_ord, pd.DataFrame): ###必须是data.frame
        print("Please provide data with dataframe type!")
        return
    if data_ord.shape[0] == 0:
        print("data is null!")
        return
    if data_ord.shape[1] != 3:
        print("Data must have three columns, please verify")
        return
    if data_ord.iloc[:,0].shape[0] != data_ord.iloc[:,1].shape[0]:
        print("ID has to be of the same length than sire and dam")
        return
    if data_ord.iloc[:,2].shape[0] != data_ord.iloc[:,1].shape[0]:
        print("sire and dam have to be of the same length")
        return
    ##如果存在na，使用sort_ped函数
    if any(data_ord.isnull().any()):
        print("There is Nan in data, first use the 'sort_ped' function")
        return
    data_ord = data_ord.astype(str)
    ###如果存在重复的行，使用sort——ped函数
    if any(data_ord.duplicated()):
        print("Duplicate in data, first use the 'sort_ped' function")
        return

    if any(data_ord.iloc[:,0].duplicated()): ##检查存在相同个体但父母却不相同的错误,即仅在个体列存在重复的情况
        print("some individuals appear more than once in the pedigree")
        return
    ###检查是否所有父母都为0
    if all(data_ord.iloc[:,1] =="0") and all(data_ord.iloc[:,2]=="0"):
        print("All dams and sires are missing")
        return
    ##检查个体同时出现在父亲列和母亲列的错误
    if any(data_ord[data_ord.iloc[:,1] != "0"].iloc[:,1].isin(data_ord[data_ord.iloc[:,1] != "0"].iloc[:,2])):
        print("Dams appearing as Sires")
        return
    ###检查一个个体本身是不是自己的父母
    if any(data_ord.iloc[:,0] == data_ord.iloc[:,1]) or any(data_ord.iloc[:,0] == data_ord.iloc[:,2]):
        print("Individual appearing as its own Sire or Dam")
        return
    N = len(data_ord.iloc[:,0])
    nPed = pd.DataFrame( {"id":pd.Series(range(N)),
                "sire":pd.Categorical(data_ord.iloc[:,1],categories=data_ord.iloc[:,0]).codes,
                "dam": pd.Categorical(data_ord.iloc[:,2],categories=data_ord.iloc[:,0]).codes
    })
    if any(nPed["id"] < nPed["sire"]) or any(nPed["id"] < nPed["dam"]):
        print("Offspring appearing before their Sires or Dams: first use the 'sort_ped' function")
        return
    nPed[nPed == -1] = N 
    f = ([0.0]*N)+[-1]
    dam = nPed["dam"]
    sire = nPed["sire"]
    dii = [0.0]*N

    dam_c =(ctypes.c_int * N) (*dam)
    sire_c = (ctypes.c_int * N) (*sire)
    f_c = (ctypes.c_double * (N+1)) (*f)
    dii_c = (ctypes.c_double * N) (*dii)
    FUNC.fcoeff(dam_c,
                    sire_c,  
                    f_c,       
                    dii_c, 
                    ctypes.c_int(N),
                    ctypes.c_int(1)
    )
    loc = pd.DataFrame({
            "x": pd.concat([nPed["id"],nPed["id"],pd.Series(range(N))]),
            "y": pd.concat([nPed["sire"],nPed["dam"],pd.Series(range(N))]),
            "data": np.append([-0.5] * (len(nPed["id"])*2),[1.0] *N)
            })
    loc=loc[loc["y"] != N]
    Tinv = sparse.coo_matrix((loc["data"], (loc["y"], loc["x"])), shape=(N, N), dtype=np.float64).tocsr()

    di =  sparse.coo_matrix((np.sqrt(1/np.array(dii_c)), (np.array(range(N)), np.array(range(N)))), shape=(N,N), dtype=np.float64).tocsr()
    #X = Tinv.dot(di).tocsc() ##upper
    X = sparse.csr_matrix(Tinv.dot(di),dtype=np.float32)

    Ip = sparse.identity(N,dtype='float32', format='dia').toarray() ## 最快的方法 7.7秒3W int8

    tu = spsolve_triangular(X,Ip,False,overwrite_A=True, overwrite_b=True)
    if Sparse:
        tu = sparse.csr_matrix(tu,dtype=np.float32)
    del Ip, X
    gc.collect()
    #tu = spsolve_triangular(X,Ip,False)
    #tu = sparse.linalg.inv(X).tocsr()
    A = tu.T.dot(tu)
    return [A, data_ord.iloc[:,0]]

def makeD(data_ord,multi=1):
    '''Calculate the additive kinship matrix using pedigree.

    data_ord: a dataframe of pedigree after sort.
    multi: int value. Default value is 1. This function uses multi-threaded calculation by default, 
        if your computer has more than one cpu, you can set the value of multi equal to the number of cpu.
    '''
    if not isinstance(multi, int):
        print("ERROR: Parameter multi should be int type!")
        return
    if multi<1:
        print('Error: multi must more than 1.')
        return
    A = makeA(data_ord,Sparse=True)
    N = A[0].shape[0]
    c =sparse.triu(A[0],format='csc')
    dA = c.diagonal()
    iAP = c.indices
    pAP = c.indptr
    xAP = c.data
    dij =  [0.0]*len(xAP)
    Di = [0]*len(iAP)
    Dp = [0]*N
    nPed = pd.DataFrame( {"id":pd.Series(range(N)),
                "sire":pd.Categorical(data_ord.iloc[:,1],categories=data_ord.iloc[:,0]).codes,
                "dam": pd.Categorical(data_ord.iloc[:,2],categories=data_ord.iloc[:,0]).codes
    })
    nPed[nPed == -1] = N 
    dam = nPed["dam"]
    sire = nPed["sire"]
    dam_c =(ctypes.c_int * N) (*dam)
    sire_c = (ctypes.c_int * N) (*sire)
    iAP_c = (ctypes.c_int * len(iAP)) (*iAP)
    pAP_c = (ctypes.c_int * len(pAP)) (*pAP)
    xAP_c = (ctypes.c_double * len(xAP)) (*xAP/2)
    
    if multi ==1:
        dij_c = (ctypes.c_double * len(xAP)) (*dij)
        Di_c = (ctypes.c_int * len(iAP)) (*Di)
        Dp_c = (ctypes.c_int * N) (*Dp)
        cnt_c = ctypes.c_int(0)
        FUNC.cald(dam_c,
            sire_c,  
            iAP_c,     
            pAP_c,	         
            xAP_c,
            ctypes.c_int(N),
            dij_c,
            Di_c,
            Dp_c,
            cnt_c
                )
        D = sparse.csr_matrix((dij_c[:cnt_c.value], Di_c[:cnt_c.value], np.append(Dp_c,cnt_c)), shape=(17, 17)).toarray()
        row, col = np.diag_indices_from(D) 
        D[row, col] = np.array(2-dA)
        return [D, data_ord.iloc[:,0]]
    else:
        global pyagh_multi_d_function
        def pyagh_multi_d_function(sub_lA):  ###listA的一部分
            lA_r = sub_lA.shape[0]
            indk_c = (ctypes.c_int * lA_r) (*sub_lA.iloc[:,0])
            indj_c = (ctypes.c_int * lA_r) (*sub_lA.iloc[:,1])
            dij =  [0.0]*lA_r
            dij_c = (ctypes.c_double * lA_r) (*dij)
            FUNC.dijp(
                dam_c,
                sire_c,
                ctypes.c_int(lA_r),
                indk_c,
                indj_c,
                iAP_c,
                pAP_c,
                xAP_c,
                dij_c
            )
            return list(dij_c),list(indk_c),list(indj_c)
        listA = pd.DataFrame({
                'Row':np.repeat(np.arange(0,len(pAP)-1),np.diff(pAP)),
                'Column':iAP
            })
        data = np.array_split(listA, multi)
        dij_list, indk_list,indj_list = [], [],[]
        with Pool() as p:
            for dij,indk,indj in p.map(pyagh_multi_d_function,data):   # or   imap()
                dij_list.extend(dij)
                indk_list.extend(indk)
                indj_list.extend(indj)
        D = sparse.coo_matrix((dij_list,(indk_list,indj_list)),shape=(N,N)).toarray()
        row, col = np.diag_indices_from(D) 
        D[row, col] = np.array(2-dA)
        return [D, data_ord.iloc[:,0]]


import pandas as pd

def sortPed(data):
    '''Sort pedigree by birthdate.

    data: the pedigree data with 3 columns(id,sire,dam) in dataframe type.
    '''
    if not isinstance(data, pd.DataFrame): ###必须是data.frame
        print("Please provide data with dataframe type!")
        return
    if data.shape[0] == 0:
        print("data is null!")
        return
    if data.shape[1] != 3:
        print("Data must have three columns, please verify")
        return
    if data.iloc[:,0].shape[0] != data.iloc[:,1].shape[0]:
        print("ID has to be of the same length than sire and dam")
        return
    if data.iloc[:,2].shape[0] != data.iloc[:,1].shape[0]:
        print("sire and dam have to be of the same length")
        return
    ###输入的数据中缺失值为NA和0，如果存在NA，转换为0
    data.fillna(0)
    ###去除完全重复的行 
    data = data.astype(str)
    data = data[~data.duplicated()]

    if any(data.iloc[:,0].duplicated()): ##检查存在相同个体但父母却不相同的错误,即仅在个体列存在重复的情况
        print("some individuals appear more than once in the pedigree")
        return
    ###检查是否所有父母都为0
    if all(data.iloc[:,1] =="0") and all(data.iloc[:,2]=="0"):
        print("All dams and sires are missing")
        return
    ##检查个体同时出现在父亲列和母亲列的错误
    if any(data[data.iloc[:,1] != "0"].iloc[:,1].isin(data[data.iloc[:,1] != "0"].iloc[:,2])):
        print("Dams appearing as Sires")
        return
    ###检查一个个体本身是不是自己的父母
    if any(data.iloc[:,0] == data.iloc[:,1]) or any(data.iloc[:,0] == data.iloc[:,2]):
        print("Individual appearing as its own Sire or Dam")
        return
    
    tmp = pd.concat([data.iloc[:,1],(data.iloc[:,2])]).drop_duplicates().astype(str)
    tmp = tmp.reset_index(drop=True)
    tmp.drop(tmp[tmp == "0"].index,inplace=True) #缺失值为0
    index = ~tmp.isin(data.iloc[:,0].astype(str))
    missingP =pd.Series(dtype=str)
    if any(index):
        missingP = tmp[index]
        lable01 = pd.concat([missingP,data.iloc[:,0]]).reset_index(drop=True)  
        sire01 = pd.concat([pd.Series(["0"]*len(missingP)),data.iloc[:,1]]).reset_index(drop=True)
        dam01 = pd.concat([pd.Series(["0"]*len(missingP)),data.iloc[:,2]]).reset_index(drop=True)
    else:
        lable01 = pd.concat([data.iloc[:,0]]).reset_index(drop=True)  
        sire01 = pd.concat([data.iloc[:,1]]).reset_index(drop=True)
        dam01 = pd.concat([data.iloc[:,2]]).reset_index(drop=True)
    sire = pd.Categorical(sire01,categories=lable01).codes
    dam = pd.Categorical(dam01,categories=lable01).codes
    nped = len(lable01)
    lable = pd.Series(range(nped))
    sire = pd.Series(sire)
    dam = pd.Series(dam)
    
    pede = pd.DataFrame({"id":lable,
                        "sire":sire,
                        "dam":dam,
                        "gene_p1": None,
                        "gene_p2":None,
                        "gene_max":None,
                        })
    noParents=(pede["sire"] == -1) & (pede["dam"] == -1)
    pede.loc[noParents,"gene_p1"] = 0
    pede.loc[noParents,"gene_p2"] = 0
    pede.loc[noParents,"gene_max"] = 0
    pede = pede.to_dict()
    
    global flag
    flag = True
    global count
    for i in range(nped):
        if flag:
            if pede["gene_p1"][i] == None:
                count = 0
                pede = getGenAncestors(pede,i)
        else:
            break
    if flag:    
        ans = pd.DataFrame({"id":lable01,
                    "sire":sire01,
                    "dam":dam01,
                    "gene_max":pd.Series(pede["gene_max"]),
                    "gene_p1":pd.Series(pede["gene_p1"]),
                    "gene_p2":pd.Series(pede["gene_p2"])
                    })
        ans['gene_min'] = ans[['gene_p1','gene_p2']].min(axis=1)
        ans.sort_values(["gene_max","gene_min"],inplace=True,ascending=[True,True])
        ans.reset_index(inplace=True,drop=True)
        return ans.iloc[:,0:3]
    else:
        print("ERROR: infinite pedigree loop involving individual")
        print("The first error individual is %s" % lable01[i-1])

def getGenAncestors(pede,i):
 
    global flag
    global count
    count += 1
 
    if count < 100:
        sire = pede["sire"][i]
        dam = pede["dam"][i]

        if sire != -1:
            tmpgenP1 = pede["gene_max"][sire]
            if tmpgenP1 == None:

                pede = getGenAncestors(pede,sire)
                genP1 = 1+ pede["gene_max"][sire]
            else:
                
                genP1 = 1 + tmpgenP1
        else:
            genP1 = 0
        if dam != -1:
            tmpgenP2 =pede["gene_max"][dam]
            if tmpgenP2 == None:
                
                pede = getGenAncestors(pede ,dam)
                genP2 = 1+ pede["gene_max"][dam]
                
            else:
                genP2 = 1+tmpgenP2
        else:
            genP2 = 0
        pede["gene_max"][i] = max(genP1,genP2) 
        pede["gene_p1"][i] = genP1
        pede["gene_p2"][i] = genP2 
        return pede
     
    else:
        print("infinite pedigree loop involving individual")
        flag = False
        return 




import os.path
import pandas as pd
def loadEgPed():
    '''Load example pedigree data.

    '''
    basepath = os.path.abspath(__file__)
    folder = os.path.dirname(basepath)
    data_path = os.path.join(folder, 'data/ped.txt')
    text = pd.read_table(data_path,header=0)

    return text
def loadEgGeno():
    '''Load example genotype data.

    '''
    basepath = os.path.abspath(__file__)
    folder = os.path.dirname(basepath)
    data_path = os.path.join(folder, 'data/geno.traw')


    return data_path 



import numpy as np
import pandas as pd
import polars as pl
import gc

## genoFile_list:两个群体合并在一起的geno文件，格式必须是traw,使用一个txt文件整合文件名  x修改支持直接输入一个文件名
## method=0 sum{[(xij - 2pi)*(xik - 2pi)] / [2pi(1-pi)]}/N as described in Yang et al. 2010 Nat Genet. 
## method=1 sum[(xij - 2pi)(xik - 2pi)] / sum[2pi(1-pi)].
## method=2 G_new 
## method=3 G_chen


def makeG(File, method, File_list = False, n1 =0, n2=0): 
    '''Calculate the additive kinship matrix using genotype.

    File: a filename of genotype in traw type of Plink. If File_list is True, this file contains the filename
        of the traw file of different chromosomes.
    method: int value of [0,1,2,3].
    method=0 sum{[(xij - 2pi)*(xik - 2pi)] / [2pi(1-pi)]}/N as described in Yang et al. 2010 Nat Genet. 
    method=1 sum[(xij - 2pi)(xik - 2pi)] / sum[2pi(1-pi)].
    method=2 G_new 
    method=3 G_chen
    n1 & n2: int value of the number of individuals of two groups. Only useful in method 2 and 3.
    File_list: bool value. Default value is False. 
    '''
    method_list =[0,1,2,3]
    if not isinstance(method, int):
        print("ERROR: Parameter method should be int type!")
        return
    if method not in method_list:
        print("ERROR: Parameter method should be in %s" %method_list)
        return
    if File_list:   
        try:
            genofile = pd.read_table(File,header=None)  ##读取包含每条染色体文件名的文件，失败报错
            for i in range(genofile.shape[0]):
                if genofile.iloc[i,0].endswith('.traw'):  ##判断是不是traw格式,并不知道文件存不存在
                    pass
                else:
                    print("ERROR: genotype file type should be .traw of PLINK")
                    return
        except Exception as reason:
            print(str(reason))
    else:
        genofile = pd.DataFrame([File])
    try:
        temp = open(genofile.iloc[0,0],"r")  
        line1 = temp.readline().split()
        ###提取出个体id，最后和G矩阵一起返回                  
        geno_id =   pd.Series(line1[6:])                  

        N = len(line1)-6 ##个体数 ##只读第一行标题 N在方法2和3中有需要
        
        G = np.zeros(N)  ###G在方法0 1 中需要

    except Exception as reason:
        print(str(reason))
    if method == 2:
        ##判断一下n1+n2=N
        if not isinstance(n1, int) or not isinstance(n2, int):
            print("ERROR: Parameter n1 and n2 should be a int")
            return
        if  n1 <= 0  or n2 <=0:
            print("ERROR: Parameter n1 and n2 should be much than 0")
            return 
        
        if n1+n2 != N:
            print("ERROR: Parameter n1 + n2 is not equal to the number of indivadual. ")
            return 

        G11=np.zeros((n1,n1))
        G12=np.zeros((n1,n2))
        G22=np.zeros((n2,n2))
        sum2pq_xx=0
        sum2pq_tb=0
        for i in range(genofile.shape[0]):
            chr_file = genofile.iloc[i,0]
            try:
                all_gen = pl.read_csv(chr_file,sep="\t",null_values="NA")
            except Exception as reason:
                print(reason)
                break
            ###这里的Z矩阵是和文献相比转置的，行是snp，列是id
            xx_gen = all_gen[:,6:(n1+6)]
            tb_gen = all_gen[:,(n1+6):]
            ###检查是否有空值，再加一个检查是不是某一个snp全是空值，全是空值的话，没法补
            if xx_gen.null_count().sum(axis=1)[0] > 0:
                print("Warning: there are mising values in first population, and imputing markers with mean value.")
                xx_gen = xx_gen.fill_null(xx_gen.mean(axis=1))
                if xx_gen.null_count().sum(axis=1)[0] > 0:
                    print("Warning: Some SNPS are missing in all individuals in first population, and imputing markers with 0.")
                    print("Warning: It is recommended that impute or filter the data first.")
                    xx_gen = xx_gen.fill_null(0)   
            if tb_gen.null_count().sum(axis=1)[0] > 0:
                print("Warning: there are mising values in second population, and imputing markers with mean value.")
                tb_gen = tb_gen.fill_null(tb_gen.mean(axis=1))
                if tb_gen.null_count().sum(axis=1)[0] > 0:
                    print("Warning: Some SNPS are missing in all individuals in second population, and imputing markers with 0.")
                    print("Warning: It is recommended that impute or filter the data first.")
                    tb_gen = tb_gen.fill_null(0)   

            p_xx = xx_gen.mean(axis=1)/2
            q_xx = 1- p_xx
            temp_2pq_xx = (2 * p_xx * q_xx).sum() #每一个snp的2pq

            p_tb = tb_gen.mean(axis=1)/2
            q_tb = 1- p_tb
            temp_2pq_tb = (2 * p_tb * q_tb).sum()
        
            z_xx = (xx_gen - 2*p_xx).to_numpy()
            z_tb = (tb_gen - 2*p_tb).to_numpy()
            G22_temp = z_tb.T.dot(z_tb)
            G12_temp = z_xx.T.dot(z_tb)
            G11_temp = z_xx.T.dot(z_xx)
            sum2pq_xx = sum2pq_xx + temp_2pq_xx
            sum2pq_tb = sum2pq_tb + temp_2pq_tb
            
            G11 = G11 + G11_temp
            G12 = G12 + G12_temp
            G22 = G22 + G22_temp
            #print(i)
            del all_gen, xx_gen, tb_gen, z_tb, z_xx
            gc.collect()
        G11_new = G11/sum2pq_xx
        G12_new = G12/(sum2pq_tb ** 0.5 * sum2pq_xx ** 0.5)
        G21_new = G12_new.T
        G22_new = G22/sum2pq_tb
        gmat = np.vstack((np.hstack((G11_new,G12_new)),np.hstack((G21_new,G22_new))))
        #np.savetxt("G_new_wgs.txt",gmat,fmt="%.10f")
        return [gmat,geno_id]
    if method == 0:
        M = 0
        for i in range(genofile.shape[0]):
            chr_file = genofile.iloc[i,0]
            try:
                all_gen = pl.read_csv(chr_file,sep="\t",null_values="NA")
            except Exception as reason:
                print(reason)
                break
            gen = all_gen[:,6:]
            if gen.null_count().sum(axis=1)[0] > 0:
                print("Warning: there are mising values in data, and imputing markers with mean value.")
                gen = gen.fill_null(gen.mean(axis=1))
                if gen.null_count().sum(axis=1)[0] > 0:
                    print("Warning: Some SNPS are missing in all individuals, and imputing markers with 0.")
                    print("Warning: It is recommended that impute or filter the data first.")
                    gen = gen.fill_null(0)  
            p = gen.mean(axis=1)/2
            q = 1-p
            dii =1/ (2 * p * q)
            Z = (gen - 2 * p).to_numpy()
            ZD = Z.T * dii.to_numpy()
            G_temp = ZD.dot(Z)
            G = G + G_temp
            M = M + gen.shape[0]
            del all_gen,gen, dii, ZD, Z, G_temp
            gc.collect()
        G = G/M
        return [G,geno_id]
    if method == 1:
        sum2pq = 0
        for i in range(genofile.shape[0]):
            chr_file = genofile.iloc[i,0]
            try:
                all_gen = pl.read_csv(chr_file,sep="\t",null_values="NA")
            except Exception as reason:
                print(reason)
                break
            gen = all_gen[:,6:]
            if gen.null_count().sum(axis=1)[0] > 0:
                print("Warning: there are mising values in data, and imputing markers with mean value.")
                gen = gen.fill_null(gen.mean(axis=1))
                if gen.null_count().sum(axis=1)[0] > 0:
                    print("Warning: Some SNPS are missing in all individuals, and imputing markers with 0.")
                    print("Warning: It is recommended that impute or filter the data first.")
                    gen = gen.fill_null(0)  
            p = gen.mean(axis=1)/2
            q = 1-p
            temp_2pq = (2 * p * q).sum()
            Z = (gen - 2*p).to_numpy()
            G_temp = Z.T.dot(Z)
            G = G + G_temp
            sum2pq = sum2pq + temp_2pq
            del all_gen, gen, Z, G_temp
            gc.collect()
        G = G/sum2pq
        return [G,geno_id]
    if method == 3:
        ##判断n1 和n2,  n1+n2应该等于N
        if not isinstance(n1, int) or not isinstance(n2, int):
            print("ERROR: Parameter n1 and n2 should be a int")
            return
        if  n1 <= 0  or n2 <=0:
            print("ERROR: Parameter n1 and n2 should be much than 0")
            return 
        if n1+n2 != N:
            print("ERROR: Parameter n1 + n2 is not equal to the number of indivadual. ")
            return 
        G11=np.zeros((n1,n1))
        G12=np.zeros((n1,n2))
        G22=np.zeros((n2,n2))
        sum2pq_xx=0
        sum2pq_tb=0
        xxbytb = 0
        for i in range(genofile.shape[0]):
            chr_file = genofile.iloc[i,0]
            try:
                all_gen = pl.read_csv(chr_file,sep="\t",null_values="NA")
            except Exception as reason:
                print(reason)
                break
            ###这里的Z矩阵是和文献相比转置的，行是snp，列是id
            xx_gen = all_gen[:,6:(n1+6)]
            tb_gen = all_gen[:,(n1+6):]
            if xx_gen.null_count().sum(axis=1)[0] > 0:
                print("Warning: there are mising values in first population, and imputing markers with mean value.")
                xx_gen = xx_gen.fill_null(xx_gen.mean(axis=1))
                if xx_gen.null_count().sum(axis=1)[0] > 0:
                    print("Warning: Some SNPS are missing in all individuals in first population, and imputing markers with 0.")
                    print("Warning: It is recommended that impute or filter the data first.")
                    xx_gen = xx_gen.fill_null(0)   
            if tb_gen.null_count().sum(axis=1)[0] > 0:
                print("Warning: there are mising values in second population, and imputing markers with mean value.")
                tb_gen = tb_gen.fill_null(tb_gen.mean(axis=1))
                if tb_gen.null_count().sum(axis=1)[0] > 0:
                    print("Warning: Some SNPS are missing in all individuals in second population, and imputing markers with 0.")
                    print("Warning: It is recommended that impute or filter the data first.")
                    tb_gen = tb_gen.fill_null(0)
            p_xx = xx_gen.mean(axis=1)/2
            q_xx = 1- p_xx
            xx_gen_list = 2 * p_xx * q_xx #每一个snp的2pq

            p_tb = tb_gen.mean(axis=1)/2
            q_tb = 1- p_tb
            tb_gen_list = 2 * p_tb * q_tb #每一个snp的2pq

            temp_2pq_xx = xx_gen_list.sum()  ##2pq的和
            temp_2pq_tb = tb_gen_list.sum()
            temp_xxbytb = ((xx_gen_list * tb_gen_list) ** 0.5).sum()  ##
            
            z_xx = (xx_gen - 2 * p_xx).to_numpy()
            z_tb = (tb_gen - 2 * p_tb).to_numpy()
            
            G11_temp = z_xx.T.dot(z_xx)
            G12_temp = z_xx.T.dot(z_tb)
            G22_temp = z_tb.T.dot(z_tb)
            
            sum2pq_xx = sum2pq_xx + temp_2pq_xx
            sum2pq_tb = sum2pq_tb + temp_2pq_tb
            xxbytb = xxbytb + temp_xxbytb
            G11 = G11 + G11_temp
            G12 = G12 + G12_temp
            G22 = G22 + G22_temp                                                                                                                                       
            #print(i)
            del all_gen, xx_gen, tb_gen, z_tb, z_xx
            gc.collect()
        G11_chen = G11/sum2pq_xx
        G12_chen = G12/xxbytb
        G21_chen = G12_chen.T
        G22_chen = G22/sum2pq_tb
        gmat = np.vstack((np.hstack((G11_chen,G12_chen)),np.hstack((G21_chen,G22_chen))))
        #np.savetxt("G_new_wgs.txt",gmat,fmt="%.10f")
        return [gmat,geno_id]


###返回亲缘系数和近交系数
import numpy as np
import pandas as pd

def coefInbreeding(A):
    '''Calculate the inbreeding coefficients using kinship matrix.

    A: a list with two elements, A[0] is kinship matrix; A[1] is id series
    '''
    if not isinstance(A, list):
        print("ERROR: Parameter A should be a list!")
        return
    if len(A) != 2:
        print("ERROR: A should have 2 elements")
        return
    if not isinstance(A[1], pd.Series):
        print("ERROR: A should have 2 elements with numpy ndarray and pandas Series")
        return
    if not isinstance(A[0], np.ndarray):
        print("ERROR: A should have 2 elements with numpy ndarray and pandas Series")
        return
    if np.isnan(A[0]).any():
        print("ERROR: Nan in A")
        return
    if A[0].shape[0] != len(A[1]):
        print("ERROR: The dimension of A is not equal to the number of individual with genotyped")
        return
 
    coef = pd.DataFrame({'ID':list(A[1]),
                'F':list(np.round(np.diag(A[0])-1,6))
           })
    return coef

def getLowerTriangularIndices(n):
    return [list(x) for x in np.triu_indices(n)]
def coefKinship(A):
    '''Calculate the relationship coefficients using kinship matrix.

    A: a list with two elements, A[0] is kinship matrix; A[1] is id series
    '''
    if not isinstance(A, list):
        print("ERROR: Parameter A should be a list!")
        return
    if len(A) != 2:
        print("ERROR: A should have 2 elements")
        return
    if not isinstance(A[1], pd.Series):
        print("ERROR: A should have 2 elements with numpy ndarray and pandas Series")
        return
    if not isinstance(A[0], np.ndarray):
        print("ERROR: A should have 2 elements with numpy ndarray and pandas Series")
        return
    if np.isnan(A[0]).any():
        print("ERROR: Nan in A")
        return
    if A[0].shape[0] != len(A[1]):
        print("ERROR: The dimension of A is not equal to the number of individual with genotyped")
        return
    n = A[0].shape[0]
    mat = np.zeros((n,n))
    for i in range(0,n):
        for j in range(i,n):
            mat[i,j] = A[0][i,j]/(A[0][i,i]*A[0][j,j])**0.5

    mat = np.round(mat,6)
    name_index = getLowerTriangularIndices(n)
    inbreed = pd.DataFrame({'ID1':list(A[1][np.array(name_index[0])]),
        'ID2':list(A[1][np.array(name_index[1])]),
        'r':list(mat[np.triu_indices(n)]),
        
        })
    return inbreed



import pandas as pd
import numpy as np
from numbers import Number
import sympy

def makeH(G,A,w=0.05):
    '''Calculate the additive kinship matrix using genotype and pedigree.

    G: a list with two elements, G[0] is kinship matrix based on genotype; G[1] is id series of genotyped individuals.
    A: a list with two elements, A[0] is kinship matrix based on pedigree; A[1] is id series of all individuals.
    w: Default value is 0.05. The weights of the G-matrix and the A-matrix
    '''

    ##提供的参数对不对 列表里包括两个元素，第一个是np矩阵，第二个是series，两者的长度要一致 w是一个数字
    ##A[0],G[0]不能有Nan9
    if not isinstance(G, list):
        print("ERROR: Parameter G should be a list!")
        return
    if len(G) != 2:
        print("ERROR: G should have 2 elements")
        return
    if not isinstance(G[1], pd.Series):
        print("ERROR: G should have 2 elements with numpy ndarray and pandas Series")
        return
    if not isinstance(G[0], np.ndarray):
        print("ERROR: G should have 2 elements with numpy ndarray and pandas Series")
        return
    if np.isnan(G[0]).any():
        print("ERROR: Nan in G")
        return
    if G[0].shape[0] != len(G[1]):
        print("ERROR: The dimension of G is not equal to the number of individual with genotyped")
        return
    ####A
    if not isinstance(A, list):
        print("ERROR: Parameter A should be a list!")
        return
    if len(A) != 2:
        print("ERROR: A should have 2 elements")
        return
    if not isinstance(A[1], pd.Series):
        print("ERROR: A should have 2 elements with numpy ndarray and pandas Series")
        return
    if not isinstance(A[0], np.ndarray):
        print("ERROR: A should have 2 elements with numpy ndarray and pandas Series")
        return
    if np.isnan(A[0]).any():
        print("ERROR: Nan in A")
        return
    if A[0].shape[0] != len(A[1]):
        print("ERROR: The dimension of A is not equal to the number of individual with genotyped")
        return
        ## W
    if not isinstance(w, Number):
        print("ERROR: Parameter w should be a number")
        return
    
    if w<0 or w>1:
        print("ERROR: Parameter w should between 0 and 1")
        return
    
    ##要检查一下测序个体是不是都在系谱
    A[1] = A[1].astype(str)
    G[1] = G[1].astype(str)
    if not all(G[1].isin(A[1])):
        print("ERROR: not all individuals with genotyped are in A matrix")
        return

    geno_id_loc = A[1][A[1].isin(G[1])] 
    index_geno = geno_id_loc.iloc[list(map(geno_id_loc.tolist().index,G[1]))].index ###按照提供的测序个体的顺序提取的在A矩阵中的index
    index_nogeno = A[1][~A[1].isin(G[1])].index ##剩下的非测序个体的index
    ##提取子集
    A11 = A[0][index_nogeno,:][:,index_nogeno]
    A12 = A[0][index_nogeno,:][:,index_geno]
    A21 = A[0][index_geno,:][:,index_nogeno]
    A22 = A[0][index_geno,:][:,index_geno]
    iA22 = np.linalg.inv(A22)
    ave_dia_G = np.trace(G[0])/G[0].shape[0]
    ave_offdia_G = (np.sum(G[0])-np.trace(G[0]))/(G[0].shape[0]*G[0].shape[0]-G[0].shape[0])
    ave_dia_A22 = np.trace(A22)/A22.shape[0]
    ave_offdia_A22 = (np.sum(A22)-np.trace(A22))/(A22.shape[0]*A22.shape[0]-A22.shape[0])
    a = sympy.Symbol("a")
    b = sympy.Symbol("b")
    ab = sympy.solve([ave_dia_G*b+a-ave_dia_A22,ave_offdia_G*b+a-ave_offdia_A22],[a,b])

    G_star = ab[a]+ab[b]*G[0]

    G_w = (1-w) * G_star + w * A22
    G_w =np.around(G_w.astype(float),6)

    H11 = A11+A12.dot(iA22).dot(G_w-A22).dot(iA22).dot(A21)
    H12 = A12.dot(iA22).dot(G_w)
    H21 = G_w.dot(iA22).dot(A21)
    H22 = G_w
    H = np.round(np.hstack((np.vstack((H11,H21)),np.vstack((H12,H22)))),2)
    return [H,pd.concat([A[1][index_nogeno], A[1][index_geno]]).reset_index(drop=True)]
    


##显性效应怎么标准化？

import numpy as np
#import math
import multiprocessing
import _thread
import numba as nb 

''' @nb.njit(parallel=True)
def numba_(a, shf):
    b = np.empty_like(a)
    rows_num = a.shape[0]
    cols_num = a.shape[1]
    for i in nb.prange(rows_num):
        b[i, shf:] = a[i, :cols_num - shf]
        b[i, :shf] = a[i, cols_num - shf:]
    return b '''

class Sum: #again, this class is from ParallelPython's example code (I modified for an array and added comments)
    def __init__(self,geno):
        self.value = np.zeros((geno.shape[0],geno.shape[0])) #this is the initialization of the sum
        self.lock = _thread.allocate_lock()
        self.count = 0

    def add(self,value):
        self.count += 1
        self.lock.acquire() #lock so sum is correct if two processes return at same time
        self.value += value #the actual summation
        self.lock.release()
''' def makematrix(geno,start,end,method):
    if method == 'dd':
        geno[geno==2] = 0  ### W矩阵
        zz11 = np.zeros((geno.shape[0],geno.shape[0]))
        for i in range(start,end):
            zz00 = geno * numba_(geno, i) 
            zz11 += (zz00.dot(zz00.T)).astype(np.float64)   ##转置相乘float32就够，但是下一步加在一起，可能几十万或者几百万会影响精度
            print(i)
    if method =='aa':
        zz11 = np.zeros((geno.shape[0],geno.shape[0]))
        for i in range(start,end):
            zz00 = geno * numba_(geno, i) 
            zz11 += (zz00.dot(zz00.T)).astype(np.float64)   ##转置相乘float32就够，但是下一步加在一起，可能几十万或者几百万会影响精度
            print(i)
    if method =='ad':
        Z = geno  ###Z矩阵
        geno[geno==2] = 0  ### W矩阵
        zz11 = np.zeros((geno.shape[0],geno.shape[0]))
        for i in range(start,end):
            Z[:,-i:] = geno[:,-i:] 
            W = numba_(geno, i)
            W[:,-i:] = Z[:,:i]
            zz00 = Z * W
            zz11 += (zz00.dot(zz00.T)).astype(np.float64)   ##转置相乘float32就够，但是下一步加在一起，可能几十万或者几百万会影响精度
            print(i)
    if method =='da':
        Z = geno  ###Z矩阵
        geno[geno==2] = 0  ### W矩阵
        zz11 = np.zeros((geno.shape[0],geno.shape[0]))
        for i in range(start,end):
            geno[:,-i:] = Z[:,-i:] 
            Z_t = numba_(Z, i)
            Z_t[:,-i:] = geno[:,:i]
            zz00 = geno * Z_t
            zz11 += (zz00.dot(zz00.T)).astype(np.float64)   ##转置相乘float32就够，但是下一步加在一起，可能几十万或者几百万会影响精度
            print(i)
    return zz11 '''
import numba as nb 
@nb.njit(parallel=True)
def one_matrix(geno,start,end,step):
    zz11 = np.zeros((geno.shape[0],geno.shape[0]))
    for i in range(start,end,step):
        zz00 = geno[:,:-i] * geno[:,i:]
        zz11 += (zz00.dot(zz00.T)).astype(np.float64)   ##转置相乘float32就够，但是下一步加在一起，可能几十万或者几百万会影响精度

    return zz11

@nb.njit(parallel=True)
def ad_matrix(geno,Z,start,end,step):
    zz11 = np.zeros((geno.shape[0],geno.shape[0]))
    for i in range(start,end,step):
        zz00 = Z[:,:-i] * geno[:,i:]
        zz11 += (zz00.dot(zz00.T)).astype(np.float64)   ##转置相乘float32就够，但是下一步加在一起，可能几十万或者几百万会影响精度

    return zz11
    
@nb.njit(parallel=True)
def da_matrix(geno,Z,start,end,step):

    zz11 = np.zeros((geno.shape[0],geno.shape[0]))
    for i in range(start,end,step):
        zz00 =geno[:,:-i] * Z[:,i:]
        zz11 += (zz00.dot(zz00.T)).astype(np.float64)   ##转置相乘float32就够，但是下一步加在一起，可能几十万或者几百万会影响精度

    return zz11

def makematrix(geno,start,end,method,step):
    if method == 'dd':
        geno[geno==2] = 0  ### W矩阵
        zz11 = one_matrix(geno,start,end,step)
    if method =='aa':
        zz11 = one_matrix(geno,start,end,step)
    if method =='ad':
        Z = geno  ###Z矩阵
        geno[geno==2] = 0  ### W矩阵
        zz11 = ad_matrix(geno,Z,start,end,step)
    if method =='da':
        Z = geno  ###Z矩阵
        geno[geno==2] = 0  ### W矩阵
        zz11 = da_matrix(geno,Z,start,end,step)
    return zz11


def makeG_inter(geno,method,multi=1):
    '''Calculate the epistatic kinship matrix using genotype.

    geno: a numpy matrix of genotype which code as 0,1 and 2. Rows are individuals and columns are SNPs.
    method: str value of 'dd','aa','ad' or 'da'.
    multi: int value. Default value is 1. This function uses multi-threaded calculation by default, 
        if your computer has more than one cpu, you can set the value of multi equal to the number of cpu.
    '''
    method_list =['dd','aa','ad','da']
    if method not in method_list:
        print("ERROR: Parameter method should be in %s" %method_list)
        return
    ###输入的geno,行是个体，列是snp，编码为012
    if not isinstance(geno, np.ndarray):
        print("ERROR: geno data should be numpy ndarray type")
        return
    if np.isnan(geno).any():
        print("ERROR: Nan in geno")
        return
    if not isinstance(multi, int):
        print("ERROR: Parameter multi should be int type!")
        return
    if multi<1:
        print('Error: multi must more than 1.')
        return
    geno = geno.astype(np.float32)
    pool = multiprocessing.Pool(processes=multi)
    sumArr = Sum(geno) #create an instance of callback class and zero the sum
    #num = math.ceil(geno.shape[1]/2)-1
    #index = np.linspace(1,num+1,multi+1,dtype=int)
    for x in range(multi):
        start= x+1 
        end = geno.shape[1]

        singlepoolresult = pool.apply_async(makematrix,(geno,start,end,method,multi),callback=sumArr.add)

    pool.close()
    pool.join() #waits for all the processes to finish
    '''###加一个检测，如果是偶数，最后多加一次平分的
    ###还要在除以对角线均值
    if (geno.shape[1] % 2) == 0: ##偶数
        if method == 'aa':
            geno1,geno2 = np.hsplit(geno,2)
            zz00 = geno1 *  geno2
            zz11 = (zz00.dot(zz00.T)).astype(np.float64) 
            sumArr.value += zz11
        if method == 'dd':
            geno[geno==2] = 0
            geno1,geno2 = np.hsplit(geno,2)
            zz00 = geno1 *  geno2
            zz11 = (zz00.dot(zz00.T)).astype(np.float64) 
            sumArr.value += zz11
        if method == 'ad':
            Z = geno  ###Z矩阵
            geno[geno==2] = 0 
            mid = int(geno.shape[1]/2)
            zz00 = Z[:,:mid] * geno[:,mid:]
            zz11 = (zz00.dot(zz00.T)).astype(np.float64) 
            sumArr.value += zz11
        if method == 'da':
            Z = geno  ###Z矩阵
            geno[geno==2] = 0 
            mid = int(geno.shape[1]/2)
            zz00 = geno[:,:mid] * Z[:,mid:]
            zz11 = (zz00.dot(zz00.T)).astype(np.float64) 
            sumArr.value += zz11'''
    M = sumArr.value/sumArr.value.diagonal().mean()
    return M


from setuptools import setup,find_packages
import os


__VERSION__ = '0.2.28'
BASE_DIR = os.path.dirname(os.path.abspath(__file__))  ##
#BASE_DIR = os.path.dirname(__file__)
os.chdir(BASE_DIR)

ext_modules = []
try:
    from pybind11.setup_helpers import Pybind11Extension 
    import pybind11
    ext_modules=[Pybind11Extension(name='FUNC',  # 模块名称
                            sources=['scrc/function_c.cpp'],    # 源码
                            define_macros = [('VERSION_INFO', __VERSION__)],
                            language='c++',
                            cxx_std=11,
                            include_dirs = [pybind11.get_include()],
                           )]

except ImportError:
    pass

setup(

    name='PyAGH',  # package name
    version= __VERSION__,  # package version
    description='A package for calculating kinship matrix',  # package description
    author='Zhao Wei',
    author_email='852322127@qq.com',
    packages=find_packages(),
    zip_safe=False,
    python_requires='>=3.5  ',
    setup_requires = ["pybind11 >= 2.9.2"],
    install_requires=["numpy >= 1.19.0",
                      "pandas >=1.1.0",
                      "sympy",
                      "scipy >= 1.7.3",
                      "polars >= 0.13.51",
                      "pybind11>= 2.9.2",
                      "matplotlib",
                      "numba"
                      ],
    include_package_data=True,
    license='MIT License',
    platforms=["all"],
    classifiers=[
          'Development Status :: 3 - Alpha',
          'Natural Language :: English',
          'License :: OSI Approved :: MIT License',
          'Intended Audience :: Science/Research',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: C++',
          'Topic :: Scientific/Engineering :: Bio-Informatics',
      ],
    ext_modules=ext_modules,
    package_data={
        '':['data/*'],
               },

)